------------------------------------------------------------------------------
--
--  package Gift_Handling (body)
--
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Text_IO;

------------------------------------------------------------------------------
--  Other packages:

with Strategy.Aux;

------------------------------------------------------------------------------

package body Gift_Handling is

   ---------------------------------------------------------------------------
   --  function Importance:
   --
   --  The value of a gift relative to the distance it has to be transported.

   function Importance (Item : in Base_Types.Gift_Type) return Float is
      use Base_Types;
      use Tracking;
      use Strategy.Aux;

      Here : constant Position_Type := Robot_Position (Own_Configuration.ID);
   begin
      return Float (Item.Weight) / Float (Metric (Here, Item.Destination));
   end Importance;

   ---------------------------------------------------------------------------
   --  function Spare_Capacity:
   --
   --  The difference between the actual and the used carrying capacity of the
   --  robot.

   function Spare_Capacity return Base_Types.Weight_Type is
      use Base_Types;

      Capacity : constant Weight_Type := Tracking.Own_Configuration.Capacity;
      Gifts    : constant Gift_Type_List := Tracking.Own_Gifts;

      Weight : Weight_Type := 0;
   begin
      for Index in Gifts'Range loop
         Weight := Weight + Gifts (Index).Weight;
      end loop;

      return Capacity - Weight;
   end Spare_Capacity;

   ---------------------------------------------------------------------------
   --  procedure Pick_Or_Drop:

   procedure Pick_Or_Drop (No_Changes :    out Boolean;
                           Command    :    out Base_Types.Command) is

      use Base_Types;

      Free : constant Weight_Type := Spare_Capacity;
      Hand : constant Gift_Type_List := Tracking.Own_Gifts;
      Tile : constant Gift_Type_List := Tracking.Get_Available_Gifts;

      Drop_Importance : array (Hand'First .. Hand'Last + 1) of Float;
      Drop_Weight     : array (Hand'First .. Hand'Last + 1) of Weight_Type;
      Pick_Importance : array (Tile'Range) of Float;
      Pick_Weight     : array (Tile'Range) of Weight_Type;
      Remaining       : array (Tile'Range) of Boolean := (others => True);

      --  Hand'Last + 1 = Don't drop a gift.
      --  Tile'Last + 1 = Don't pick up a gift.

   begin --  Pick_Or_Drop
      for Index in Hand'Range loop
         Drop_Importance (Index) := Importance (Hand (Index));
         Drop_Weight (Index) := Hand (Index).Weight;
      end loop;
      Drop_Importance (Drop_Importance'Last) := 0.0;
      Drop_Weight (Drop_Weight'Last) := 0;

      for Index in Tile'Range loop
         Pick_Importance (Index) := Importance (Tile (Index));
         Pick_Weight (Index) := Tile (Index).Weight;
      end loop;

      declare
         Effective_Capacity        : Weight_Type;
         Importance_Growth         : Float := 0.0;
         Drop                      : Gift_Counter := Hand'Last + 1;
         Pick                      : Gift_Counter := Tile'Last + 1;
         Found_One_More_To_Pick_Up : Boolean;
      begin
         for Drop_Index in Drop_Importance'Range loop
            Effective_Capacity := Free + Drop_Weight (Drop_Index);
            for Pick_Index in Pick_Importance'Range loop
               if Effective_Capacity < Pick_Weight (Pick_Index) then
                  null; --  Too heavy.
               elsif Pick_Importance (Pick_Index)
                   - Drop_Importance (Drop_Index) > Importance_Growth then
                  Drop := Drop_Index;
                  Pick := Pick_Index;
                  Importance_Growth := Pick_Importance (Pick_Index) 
                                         - Drop_Importance (Drop_Index);
               end if;
            end loop;
         end loop;

         if Drop in Hand'Range then
            No_Changes := False;
            Command := (Action     => Base_Types.Drop,
                        Gift_Count => 1,
                        Gifts      => (1 => Hand (Drop).ID));
         elsif Pick in Tile'Range then
            No_Changes := False;
            Command := (Action     => Base_Types.Pick,
                        Gift_Count => 1,
                        Gifts      => (1 => Tile (Pick).ID));
            Remaining (Pick) := False;
            Effective_Capacity := Effective_Capacity - Pick_Weight (Pick);

            loop
               Found_One_More_To_Pick_Up := False;
               Importance_Growth := 0.0;

               for Index in Remaining'Range loop
                  Remaining (Index) :=
                    Remaining (Index) and then
                    Effective_Capacity >= Pick_Weight (Index);

                  if Remaining (Index) and
                     Pick_Importance (Index) > Importance_Growth then

                     Found_One_More_To_Pick_Up := True;
                     Importance_Growth := Pick_Importance (Index);
                     Pick := Index;
                  end if;
               end loop;

               if Found_One_More_To_Pick_Up then
                  Command := (Action     => Base_Types.Pick,
                              Gift_Count => Command.Gift_Count + 1,
                              Gifts      => Command.Gifts &
                                            (1 => Tile (Pick).ID));
                  Remaining (Pick) := False;
                  Effective_Capacity := Effective_Capacity - Pick_Weight (Pick);
               else
                  exit;
               end if;
            end loop;
         else
            No_Changes := True;
         end if;
      end;
   end Pick_Or_Drop;

   ---------------------------------------------------------------------------

end Gift_Handling;
