------------------------------------------------------------------------------
--
--  package Strategy (body)
--
------------------------------------------------------------------------------
--  Standard packages:

with Ada.Text_IO;

------------------------------------------------------------------------------
--  Other packages:

with Landscapes;
with Tracking;
with Gift_Handling;
with Zero_Filled_Image;
with Bids;

------------------------------------------------------------------------------

package body Strategy is

   ---------------------------------------------------------------------------
   --  Delivery and source (home base) landscapes:

   Deliveries : Landscapes.Landscape_Reference;
   Sources    : Landscapes.Landscape_Reference;

   Source_Map : constant Landscapes.Mapping :=
     (Base_Types.Plain => (Interpolation => Landscapes.Flexible,
                           Height        => 0.0),
      Base_Types.Water => (Interpolation => Landscapes.Ignore),
      --Base_Types.Water => (Interpolation => Landscapes.Fixed,
      --                     Height        => -2.0),
      Base_Types.Wall  => (Interpolation => Landscapes.Ignore),
      Base_Types.Home  => (Interpolation => Landscapes.Fixed,
                           Height        => 10.0));

   ---------------------------------------------------------------------------
   --  procedure Initialise:

   procedure Initialise (Board : in     Base_Types.Board_Type) is
      use Landscapes;
   begin --  Initialise
      Deliveries := new Landscape (Board'Range (1),
                                   Board'Range (2));
      To_Landscape (Source => Board,
                    Target => Deliveries.all);
      Sources := new Landscape (Board'Range (1),
                                Board'Range (2));
      To_Landscape (Source => Board,
                    Target => Sources.all,
                    Map    => Source_Map);
   end Initialise;

   ---------------------------------------------------------------------------
   --  procedure Choose_Move:

   Image_Number : Natural := 0;

   procedure Choose_Move (Bid  :    out Base_Types.Bid_Type;
                          Move :    out Base_Types.Command) is
      use Base_Types;
      use Landscapes;
      use Tracking;

      Current_Position : constant Position_Type :=
        Robot_Position (Own_Configuration.ID);
      Board : constant Board_Type := Return_Board_State;
   begin --  Choose_Move

  Should_Drop:
      declare
         Gifts : constant Gift_Type_List := Own_Gifts;
         Number_To_Be_Dropped : Gift_Counter := 0;
         Should_Be_Dropped : array (Gifts'Range) of Boolean;
      begin
         for Index in Gifts'Range loop
            Should_Be_Dropped (Index) :=
              Gifts (Index).Destination = Current_Position;

            if Should_Be_Dropped (Index) then
               Number_To_Be_Dropped := Number_To_Be_Dropped + 1;
            end if;
         end loop;
         if Number_To_Be_Dropped > 0 then
            Bid := Bids.Bid (Circumstances =>
                               (Situation            => Bids.Should_Drop,
                                Number_To_Be_Dropped => Number_To_Be_Dropped));
            Move := (Action     => Drop,
                     Gift_Count => Number_To_Be_Dropped,
                     Gifts      => (others => 0));
            for Index in Gifts'Range loop
               if Should_Be_Dropped (Index) then
                  Move.Gifts (Number_To_Be_Dropped) := Gifts (Index).ID;
                  Number_To_Be_Dropped := Number_To_Be_Dropped - 1;
               end if;
            end loop;
            return;
         end if;
      end Should_Drop;

  Could_Push:
      declare
         Pushed_Position       : Position_Type;
         Pushables             : Natural;
         Max_Pushables         : Natural := 0;
         Max_Pushables_Bearing : Directions;
      begin
         for Bearing in Directions loop
            Pushed_Position := Translate (Current_Position, Bearing);
            Pushables := 0;
            while Tracking.Board (Pushed_Position).Occupied loop
               Pushables := Pushables + 1;
               Translate (Pushed_Position, Bearing);
            end loop;
            if Tracking.Board (Pushed_Position).Tile = Water
              and Pushables > Max_Pushables then
               Max_Pushables := Pushables;
               Max_Pushables_Bearing := Bearing;
            end if;
         end loop;

         if Max_Pushables > 0 then
            Bid := Bids.Bid(Circumstances => (Situation     => Bids.Could_Push,
                                              Max_Pushables => Max_Pushables));
            Move := (Action     => Base_Types.Move,
                     Gift_Count => 0,
                     Direction  => Max_Pushables_Bearing);
            return;
         end if;
      end Could_Push;

  Should_Pick:
      declare
         use Gift_Handling;
         Continue : Boolean;
      begin
         Pick_Or_Drop (No_Changes => Continue,
                       Command    => Move);
         if not Continue then
            Bid := Bids.Bid (Circumstances
                             => (Situation           => Bids.Should_Pick,
                                 Number_To_Be_Picked => Move.Gift_Count));
            return;
         end if;
      end Should_Pick;

   Update_Delivery_Landscape:
      declare

         Gifts : constant Gift_Type_List := Own_Gifts;
      begin
         for X in Deliveries'Range(1) loop
            for Y in Deliveries'Range(2) loop
               if Deliveries (X, Y).Interpolation = Fixed then
                  Deliveries (X, Y) :=
                    (Interpolation => Flexible,
                     Height        => Deliveries (X, Y).Height);
               end if;
            end loop;
         end loop;

         for Index in Gifts'Range loop
            declare
               X      : X_Coordinate renames Gifts (Index).Destination.X;
               Y      : Y_Coordinate renames Gifts (Index).Destination.Y;
               Weight : Weight_Type renames Gifts (Index).Weight;
            begin
               if Deliveries (X, Y).Interpolation = Fixed then
                  Deliveries (X, Y).Height :=
                    Deliveries (X, Y).Height + Float (Weight);
               else
                  Deliveries (X, Y) := (Interpolation => Fixed,
                                        Height        => Float (Weight));
               end if;
            end;
         end loop;

         for X in Deliveries'Range(1) loop
            for Y in Deliveries'Range(2) loop
               if X in Board'Range (1) and Y in Board'Range (2) then
                  case Board (X, Y).Tile is
                     when Plain =>
                        null;
                     when Water =>
                        Deliveries (X, Y) := (Interpolation => Ignore);
                        --Deliveries (X, Y) := (Interpolation => Fixed,
                        --                      Height        => -10.0);
                     when Wall =>
                        Deliveries (X, Y) := (Interpolation => Ignore);
                     when Home =>
                        null;
                  end case;
               end if;
            end loop;
         end loop;

         Deliveries (Current_Position.X, Current_Position.Y) :=
           (Interpolation => Fixed,
            Height        => 0.0);
         Interpolate (Item       => Deliveries.all,
                      Iterations => 20);
         --Save (Item => Deliveries.all,
         --      Name => "deliveries.ppm");
      end Update_Delivery_Landscape;

   Update_Source_Landscape:
      declare
         Average_Weight : constant Float := 10.0;
         Spare_Capacity : constant Float :=
                            Float (Gift_Handling.Spare_Capacity);
      begin
         for X in Board'Range (1) loop
            for Y in Board'Range (2) loop
               if Board (X, Y).Net_Dropped_Gifts > 0 then
                  Sources (X, Y) :=
                    (Interpolation => Fixed,
                     Height        =>
                       Float'Min (Float (Board (X, Y).Net_Dropped_Gifts) *
                                    Average_Weight,
                                  Spare_Capacity));
               end if;
            end loop;
         end loop;

         Sources (Current_Position.X, Current_Position.Y) :=
           (Interpolation => Fixed,
            Height        => 0.0);
         Interpolate (Item       => Sources.all,
                      Iterations => 20);
         --Save (Item => Sources.all,
         --      Name => "sources.ppm");
         Sources (Current_Position.X, Current_Position.Y) :=
           (Interpolation => Flexible,
            Height        => -1.0);
      end Update_Source_Landscape;

   Choose_Direction:
      declare
         Total_Capacity : Weight_Type renames Own_Configuration.Capacity;
         Spare_Capacity : Weight_Type renames Gift_Handling.Spare_Capacity;

         Deliveries_Factor : constant Float := Float (Total_Capacity -
                                                      Spare_Capacity);
         Sources_Factor    : constant Float := Float (Spare_Capacity);

         New_Position    : Position_Type;
         New_Level       : Float;

         Has_Estimate        : Boolean := False;
         Best_Direction      : Directions;
         Largest_Improvement : Float;
      begin

--    Save_Composite_Landscape:
--       declare
--          function Image is
--            new Zero_Filled_Image (Num => Natural, Width => 3);
--          Composite : Landscape := Deliveries.all;
--       begin
--          for X in Composite'Range (1) loop
--             for Y in Composite'Range (2) loop
--                if Deliveries (X, Y).Interpolation = Ignore or
--                  Sources (X, Y).Interpolation = Ignore then
--                   Composite (X, Y) := (Interpolation => Ignore);
--                else
--                   Composite (X, Y).Height :=
--                     Deliveries (X, Y).Height * Deliveries_Factor +
--                     Sources (X, Y).Height * Sources_Factor;
--                end if;
--             end loop;
--          end loop;
--
--          Save (Item => Composite,
--                Name => "composite-" & Image (Image_Number) & ".ppm");
--          Image_Number := Image_Number + 1;
--       end Save_Composite_Landscape;

         for Direction in Directions loop
            New_Position := Translate (Point     => Current_Position,
                                       Direction => Direction);

            if Deliveries (New_Position.X,
                           New_Position.Y).Interpolation = Ignore or
               Sources (New_Position.X,
                        New_Position.Y).Interpolation = Ignore then
               null;
            else
               New_Level :=
                 Sources (New_Position.X, New_Position.Y).Height * Sources_Factor +
                 Deliveries (New_Position.X, New_Position.Y).Height * Deliveries_Factor;

               if Tracking.Board (New_Position).Tile = Water then
                  null;
               elsif not Has_Estimate then
                  Has_Estimate := True;
                  Best_Direction := Direction;
                  Largest_Improvement := New_Level;
               elsif New_Level > Largest_Improvement then
                  Best_Direction := Direction;
                  Largest_Improvement := New_Level;
               end if;
            end if;
         end loop;

         if Has_Estimate then
            Bid := Bids.Bid (Circumstances => (Situation => Bids.Normal));
            Move := (Action     => Base_Types.Move,
                     Direction  => Best_Direction,
                     Gift_Count => 0);
         else
            Bid := 1;
            Move := Null_Command;
         end if;
      end Choose_Direction;
   end Choose_Move;

   ---------------------------------------------------------------------------

end Strategy;
