with Hash_G;
with Ada.Text_IO;

package body Tracking is
   ----------------------------------------------------------------------------
   -- Global variables for keeping the states of the robot and the board.

   type Board_Reference is access all Base_Types.Board_Type;
   Board_State : Board_Reference;

   --Board_State : Base_Types.Board_Type; -- Should be removed, I think.

   My_Robot_Configuration : Base_Types.Robot_Configuration;

   ---------------------------------------------------------------------------
   --  Gifts carried by our robot:

   Gifts        : Base_Types.Gift_List (1 .. Base_Types.Gift_Counter'Last);
   Gift_Counter : Base_Types.Gift_Counter := 0;

   ---------------------------------------------------------------------------
   --  Gifts available to pick up from the current location of our robot:

   Available_Gifts : Base_Types.Gift_List (1 .. Base_Types.Gift_Counter'Last);
   Number_Of_Available_Gifts : Base_Types.Gift_Counter := 0;

   ---------------------------------------------------------------------------
   --  Hash table to keep track of the robots.
   --  First we instantiate the package...
   function Robot_Hash (Robot_ID : Base_Types.Robot_ID) return Natural is
      use type Base_Types.Robot_ID;
   begin
      return Natural (abs Robot_ID);
   exception
      when others =>
         Ada.Text_IO.Put_Line (File => Ada.Text_IO.Standard_Error,
                               Item => "Exception in function Robot_Hash.");
         raise;
   end Robot_Hash;

   function Compare_Robot_Type (L, R : Base_Types.Robot_Type) return Boolean is
      use type Base_Types.Robot_Type;
   begin
      return L = R;
   end Compare_Robot_Type;

   function Compare_Robot_ID (L, R : Base_Types.Robot_ID) return Boolean is
      use type Base_Types.Robot_ID;
   begin
      return L = R;
   end Compare_Robot_ID;

   package Robot_Hash_Table is
      new Hash_G (Base_Types.Robot_ID,
                  Base_Types.Robot_Type,
                  Robot_Hash,
                  Compare_Robot_ID,
                  Compare_Robot_Type);
   --  ...and then define the table
   Robot_Table : Robot_Hash_Table.Hash_Table_Type
     := Robot_Hash_Table.Create (1000);
   ---------------------------------------------------------------------------
   --  Hash table to keep track of living robots
   --  We can reuse Robot_Hash and Compare_Robot_ID

   package Living_Robot_Hash_Table is
      new Hash_G (Base_Types.Robot_ID,
                  Boolean,
                  Robot_Hash,
                  Compare_Robot_ID,
                  "=");
   Living_Robot_Table : Living_Robot_Hash_Table.Hash_Table_Type
     := Living_Robot_Hash_Table.Create (1000);
   ---------------------------------------------------------------------------
   --  Hash table to keep track of the gifts.
   --  First we instantiate the package...
   function Gift_Hash (Gift_ID : Base_Types.Gift_ID) return Natural is
   begin
      return Natural (abs Integer(Gift_ID));
   end Gift_Hash;

   function Compare_Gift_Type (L, R : Base_Types.Gift_Type) return Boolean is
      use type Base_Types.Gift_Type;
   begin
      return L = R;
   end Compare_Gift_Type;

   function Compare_Gift_ID (L, R : Base_Types.Gift_ID) return Boolean is
      use type Base_Types.Gift_ID;
   begin
      return L = R;
   end Compare_Gift_ID;

   package Gift_Hash_Table is
      new Hash_G (Base_Types.Gift_ID,
                  Base_Types.Gift_Type,
                  Gift_Hash,
                  Compare_Gift_ID,
                  Compare_Gift_Type);
   --  ...and then define the table
   Gift_Table : Gift_Hash_Table.Hash_Table_Type
     := Gift_Hash_Table.Create (10000);
   ----------------------------------------------------------------------------
   -- Setting and updating states
   procedure Set_Board (Board : in Base_Types.Board_Type) is

      use Base_Types;

   begin
      Board_State := new Board_Type'(Board);
   end Set_Board;

   procedure Set_Robot_Configuration

     (Configuration : in Base_Types.Robot_Configuration) is

   begin

      My_Robot_Configuration := Configuration;

   end Set_Robot_Configuration;

   function Robot_Position (Robot : in Base_Types.Robot_ID)
     return Base_Types.Position_Type is
      use Ada.Text_IO;
      use Base_Types;

      Robot_Info : Robot_Type;
   begin
      if Robot_Hash_Table.Key_Exists (Hash_Table => Robot_Table,
                                      Key        => Robot) then
         Robot_Info := Robot_Hash_Table.Retrieve (Hash_Table => Robot_Table,
                                                  Key        => Robot);
         return Robot_Info.Position;
      else
         Put_Line (File => Standard_Error,
                   Item => "Could not find position of robot " &
                           Robot_ID'Image (Robot) & ".");
         return (X => 0, Y => 0);
      end if;
   exception
      when others =>
         Put_Line (File => Standard_Error,
                   Item => "An exception was raised in function " &
                           "Robot_Position.");
         Put_Line (File => Standard_Error,
                   Item => "Could not find position of robot " &
                           Robot_ID'Image (Robot) & ".");
         return (X => 0, Y => 0);
   end Robot_Position;

   ---------------------------------------------------------------------------
   procedure Clear_Tile (Robot  : in Base_Types.Robot_ID;
                         Position : in Base_Types.Position_Type) is
      use Base_Types;
   begin
      if Robot = My_Robot_Configuration.ID then
         if Board_State (Position.X, Position.Y).Tile = Home then
            Board_State (Position.X, Position.Y).Tile := Plain;
         end if;
         Board_State (Position.X, Position.Y).Net_Dropped_Gifts
           := Number_Of_Available_Gifts;
      end if;
      Board_State (Position.X, Position.Y) :=
        (Occupied => False,
         Tile =>
           Board_State (Position.X, Position.Y).Tile,
         Net_Dropped_Gifts =>
           Board_State (Position.X, Position.Y).Net_Dropped_Gifts);
   end Clear_Tile;

   procedure Set_Tile (Robot  : in Base_Types.Robot_ID;
                       New_Position : in Base_Types.Position_Type) is
      use Base_Types;
   begin
      Board_State (New_Position.X, New_Position.Y) :=
        (Occupied =>
           True,
         Tile =>
           Board_State (New_Position.X, New_Position.Y).Tile,
         Net_Dropped_Gifts =>
           Board_State (New_Position.X, New_Position.Y).Net_Dropped_Gifts,
         Robot =>
           Robot);
   end Set_Tile;

   procedure Update_With (Robot  : in Base_Types.Robot_ID;
                          Result : in Server_Communications.Action_Result) is

      procedure Update_Robot_Position_In_Table
        (Direction : in Base_Types.Directions) is
         use Robot_Hash_Table;
         use Base_Types;
         Robot_Info : Base_Types.Robot_Type
           := Retrieve (Hash_Table => Robot_Table, Key => Robot);
      begin
         Base_Types.Translate (Point => Robot_Info.Position,
                               Direction => Direction);
         Modify (Hash_Table => Robot_Table,
                 Key => Robot,
                 New_Data => Robot_Info);
      end Update_Robot_Position_In_Table;

      procedure Update_Gifts is
         use Base_Types;
         use Server_Communications;
         use Robot_Hash_Table;
         Robot_Info : Base_Types.Robot_Type
           := Retrieve (Hash_Table => Robot_Table, Key => Robot);
      begin
         if Result.Action = Server_Communications.Picked_Gift_Up then
            -------------------------------------------------------------------
            -- Case: picked up gift
            Robot_Info.Gifts := Robot_Info.Gifts + 1;
            if Board_State (Robot_Info.Position.X,
                            Robot_Info.Position.Y).Net_Dropped_Gifts > 0 then
               Board_State (Robot_Info.Position.X,
                            Robot_Info.Position.Y).Net_Dropped_Gifts :=
                 Board_State (Robot_Info.Position.X,
                            Robot_Info.Position.Y).Net_Dropped_Gifts - 1;
            end if;
            if Robot = My_Robot_Configuration.ID then -- if it is our robot!
                 Gift_Counter := Gift_Counter + 1;
                 Gifts (Gift_Counter) := Result.Gift; -- add it to the list
            end if;
         else
            -------------------------------------------------------------------
            -- Case: dropped gift

            if Robot_Info.Gifts > 0 then
               Robot_Info.Gifts := Robot_Info.Gifts - 1;
            end if;

            if Gift_Hash_Table.Key_Exists (Hash_Table => Gift_Table,
                                           Key => Result.Gift) then
               -- Check if we know the package at all
            Destination_Local:
               declare
                  Known_Gift : Gift_Type := Gift_Hash_Table.Retrieve
                        (Hash_Table => Gift_Table,
                         Key => Result.Gift);
               begin
                  if Known_Gift.Destination /= Robot_Info.Position then
                     -- Update, since we _know that it didn't reach its
                     -- destination
                     Board_State (Robot_Info.Position.X,
                                  Robot_Info.Position.Y).Net_Dropped_Gifts :=
                       Board_State (Robot_Info.Position.X,
                                    Robot_Info.Position.Y).Net_Dropped_Gifts + 1;
                  end if;
               end Destination_Local;
            else  -- We didn't know the package but we assume that it didn't
               -- reach its final destination.
               Board_State (Robot_Info.Position.X,
                            Robot_Info.Position.Y).Net_Dropped_Gifts
                 := Board_State (Robot_Info.Position.X,
                                 Robot_Info.Position.Y).Net_Dropped_Gifts + 1;
            end if;

            if Robot = My_Robot_Configuration.ID then
               -- if it is our robot, we need to remove the dropped gift and do a
               -- bit of reshuffling
               declare
                  use Base_Types;
                  Already_Found : Boolean := False;
               begin
                  for Gift_Item in 1 .. Gift_Counter loop
                     if Gifts (Gift_Item) = Result.Gift then
                        Already_Found := True;
                     elsif Already_Found then
                        Gifts (Gift_Item - 1) := Gifts (Gift_Item);
                     end if;
                  end loop;
               end;
               -- finally decrement the counter.
               Gift_Counter := Gift_Counter - 1;
            end if;
         end if;
         ----------------------------------------------------------------------

         Modify (Hash_Table => Robot_Table,
                 Key        => Robot,
                 New_Data   => Robot_Info);
      end Update_Gifts;

      procedure Set_XY is
         use Robot_Hash_Table;
         use Base_Types;
         Robot_Info : Base_Types.Robot_Type;
      begin
         if Key_Exists (Hash_Table => Robot_Table,
                        Key        => Robot) then
            Robot_Info := Retrieve (Hash_Table => Robot_Table,
                                    Key => Robot);
            -- Remove from board at old position!
            Clear_Tile (Robot, Robot_Info.Position);
            -- Copy new position
            Robot_Info.Position.X := Result.X;
            Robot_Info.Position.Y := Result.Y;
            -- Update table
            Modify (Hash_Table => Robot_Table,
                    Key => Robot,
                    New_Data => Robot_Info);
            -- Insert at new position on board!!!
            Set_Tile (Robot, Robot_Info.Position);
         else
            Robot_Info := (ID => Robot,
                           Position => (X => Result.X,
                                        Y => Result.Y),
                           Gifts => 0); -- A reasonable assumption?
            Insert (Hash_Table => Robot_Table,
                    Key => Robot,
                    Data => Robot_Info);
            -- Insert at new position on board!!!
            Set_Tile (Robot, Robot_Info.Position);
         end if;
      end Set_XY;

      Ignored : Boolean; -- Dummy to make Hash_table happy!
      use Base_Types;
      use Server_Communications;
   begin
      Living_Robot_Hash_Table.Insert_Or_Modify
        (Hash_Table => Living_Robot_Table,
         Key => Robot,
         Data => True,
         Newer => Ignored);
      case Result.Action is
         when Moved_North =>
            Clear_Tile (Robot, Robot_Position (Robot));
            Set_Tile (Robot, Translate (Robot_Position (Robot), North));
            Update_Robot_Position_In_Table (North);
         when Moved_East =>
            Clear_Tile (Robot, Robot_Position (Robot));
            Set_Tile (Robot, Translate (Robot_Position (Robot), East));
            Update_Robot_Position_In_Table (East);
         when Moved_South =>
            Clear_Tile (Robot, Robot_Position (Robot));
            Set_Tile (Robot, Translate (Robot_Position (Robot), South));
            Update_Robot_Position_In_Table (South);
         when Moved_West =>
            Clear_Tile (Robot, Robot_Position (Robot));
            Set_Tile (Robot, Translate (Robot_Position (Robot), West));
            Update_Robot_Position_In_Table (West);
         when Picked_Gift_Up | Dropped_Gift =>
            Update_Gifts;
         when Set_XY =>
            Set_XY;
      end case;
   end Update_With;

   procedure Clear_Available_Gifts is
   begin
      Number_Of_Available_Gifts := 0;
   end Clear_Available_Gifts;

   procedure Gift_At_Tile (Gift : Base_Types.Gift_Type) is
      use Base_Types;
      Ignored : Boolean;
   begin
Ada.Text_IO.Put_Line ("Gift_At_Tile: ID = " & Gift.ID'Img);
      -- We collect information about all the gifts we have seen.
      Gift_Hash_Table.Insert_Or_Modify (Hash_Table => Gift_Table,
                                        Key => Gift.ID,
                                        Data => Gift,
                                        Newer => Ignored);
      -- Update the list of Available_Gifts
      Number_Of_Available_Gifts := Number_Of_Available_Gifts + 1;
      Available_Gifts (Number_Of_Available_Gifts) := Gift.ID;
   end Gift_At_Tile;

   procedure Start_Keeping_Track_Of_Robots is
   begin
      Living_Robot_Hash_Table.Clear (Living_Robot_Table);
   end Start_Keeping_Track_Of_Robots;

   procedure Stop_Keeping_Track_Of_Robots is
      procedure Clean_Up (Key      : in Base_Types.Robot_ID;
                          Data     : in out Base_Types.Robot_Type;
                          Continue : in out Boolean) is
         use type Base_Types.Position_Type;

         Temp_Robot : Base_Types.Robot_Type;
      begin
         if not Living_Robot_Hash_Table.Key_Exists
           (Hash_Table => Living_Robot_Table,
            Key => Key) then
            Temp_Robot := Robot_Hash_Table.Retrieve (Hash_Table => Robot_Table,
                                                     Key => Key);
            -- Should clean up in the table
            -- Should remove from the board (more important?)
            if Data.Position /= (X => 0, Y => 0) then
               Clear_Tile (Robot => Key, Position => Temp_Robot.Position);
               Data.Position := (X => 0, Y => 0);
            end if;
         end if;
      end Clean_Up;
      procedure Remove_Dead_Robots is new Robot_Hash_Table.Visit (Clean_Up);
   begin
      Remove_Dead_Robots (Robot_Table);  -- Kills robots
   end Stop_Keeping_Track_Of_Robots;

   procedure Register_Robot_Exists (Robot : in Base_Types.Robot_ID) is
      Ignored : Boolean; -- Dummy to satify insert_or_modify function
   begin
      Living_Robot_Hash_Table.Insert_Or_Modify
        (Hash_Table => Living_Robot_Table,
         Key => Robot,
         Data => True,
         Newer => Ignored);
   end Register_Robot_Exists;
   ----------------------------------------------------------------------------
   -- Returning information on the present state.

   function Return_Board_State return Base_Types.Board_Type is
   begin
      return Board_State.all;
   end Return_Board_State;

   function Board (Position : Base_Types.Position_Type)
     return Base_Types.Tile_State is
   begin
      return Board_State.all (Position.X, Position.Y);
   end Board;

   function Own_Gifts return Base_Types.Gift_List is
   begin
      return Gifts (1 .. Gift_Counter);
   end Own_Gifts;

   function Own_Gifts return Base_Types.Gift_Type_List is
      Temp_Gifts : Base_Types.Gift_Type_List (1 .. Gift_Counter);
   begin
      for Item in Temp_Gifts'range loop
         Temp_Gifts (Item)
           := Gift_Hash_Table.Retrieve (Hash_Table => Gift_Table,
                                        Key => Gifts (Item));
      end loop;
      return Temp_Gifts;
   end Own_Gifts;

   function Get_Available_Gifts return Base_Types.Gift_List is
   begin
      return Available_Gifts (1 .. Number_Of_Available_Gifts);
   end Get_Available_Gifts;

   function Get_Available_Gifts return Base_Types.Gift_type_List is
      Temp_Gifts : Base_Types.Gift_Type_List (1 .. Number_Of_Available_Gifts);
   begin
      for Item in Temp_Gifts'range loop
         Temp_Gifts (Item)
           := Gift_Hash_Table.Retrieve (Hash_Table => Gift_Table,
                                        Key => Available_Gifts (Item));
      end loop;
      return Temp_Gifts;
   end Get_Available_Gifts;

   function Weight (Gift : Base_Types.Gift_ID) return Base_Types.Weight_Type is
      use Ada.Text_IO;

      Current_Type : Base_Types.Gift_Type;
   begin
      Current_Type := Gift_Hash_Table.Retrieve (Hash_Table => Gift_Table,
                                                Key        => Gift);
      return Current_Type.Weight;
   exception
      when Gift_Hash_Table.Key_Not_Found_Error =>
         Put_Line (File => Standard_Error,
                   Item => "Tracking.Weight: Does not know the gift " &
                           Base_Types.Gift_ID'Image (Gift) & ".");
         raise;
   end Weight;

   function Own_Configuration return Base_Types.Robot_Configuration is
   begin
      return My_Robot_Configuration;
   end Own_Configuration;

end Tracking;
